#include <Servo.h>

#define START_BUTTON 13
#define Y_STEP 12
#define Y_DIR  11
#define Z_STEP 10
#define Z_DIR 9
#define COLOR_S0 8
#define COLOR_S1 7
#define GRIPPER 6
#define COLOR_LED 5
#define COLOR_OUT 4
#define COLOR_S3 3
#define COLOR_S2 2
#define END_Y A5
#define END_Z A4
#define FRUIT_CHOICE A1
#define DISTANCE_SENSOR A0
#define Y_ENABLE A2

int step = 5;

//Fruit selection variables
int fruitSelected;
float fruit1Max = 340; //adjust if needed (limit value of potentiometer)
float fruit2Max = 680; //adjust if needed (limit value of potentiometer)

//Object position variables
float objectStart;
float objectEnd;
bool objectFound = 0;
float groundDistance;
float distanceError = 10; //Maximum error when detecting table in mm

//Y axis variables
float yPos;
float yMax = 2300; //Max Y position in steps
float yGoal;
float stepToPosY = 1; //Could be changed to mm if wanted
float sensorOffset = 200; //offset of the distance sensor in steps
float scanOffset = 650; //debut of scanning zone
float colorReadPos = 80; // pos to go to read color
int yBins[2] = { (colorReadPos+100) , (yMax-colorReadPos)}; //pos right bin, pos left bin in steps 
bool bin = 1; //0 = right Bin, 1 = left Bin

//Z axis variables
float zPos;
int zGoal = 1000; //initial value >> to find the smallest later
float stepToPosZ = 0.01; //2mm per revolution and 200 step per revolution
float zSafeDistance = 80; //Distance from ground in mm that is good for grabbing
float dropDistance = 90; //Distance from the ground to drop the fruit from
int zObjectTransport = 5; //Height in mm that is good for transport

//Gripper variables
Servo gripper; 
float gripperAngleClosed = 105; //Angle when closed
float gripperAngleOpen = 55; //Angle when open

//Color detection variables
int objectRed;
int objectGreen;
int objectBlue;
int objectRGB[3] = {0,0,0}; 
int redMin = 40; 
int redMax = 1299; 
int greenMin = 48; 
int greenMax = 1009; 
int blueMin = 35; 
int blueMax = 1009; 
int redPW = 0;
int greenPW = 0;
int bluePW = 0;
int colorError = 20; //half interval of acceptance for object RGB value to be in OK zone
int fruitColors[3][3] = { // 2-dim array with 3 fruit choice and with 3 limit value for RGB(0-255)
  {246 , 250 , 231}, // RGB limit value fruit 1 (Apple limit)
  {143, 74, 102}, // RGB limit value fruit 2 (Peach limit)
  {234, 191, 212}, // RGB limit value fruit 3 (Strawberry limit)
};




void setup() {
  Serial.begin(9600);
  //Outputs
  pinMode(Y_STEP, OUTPUT);
  pinMode(Y_DIR, OUTPUT);
  pinMode(Z_STEP, OUTPUT);
  pinMode(Z_DIR, OUTPUT);
  pinMode(COLOR_S0, OUTPUT);
  pinMode(COLOR_S1, OUTPUT);
  pinMode(COLOR_LED,OUTPUT);
  pinMode(COLOR_S2, OUTPUT);
  pinMode(COLOR_S3, OUTPUT);
  pinMode(Y_ENABLE, OUTPUT);
  //Inputs
  pinMode(START_BUTTON, INPUT_PULLUP);
  pinMode(COLOR_OUT, INPUT);
  pinMode(END_Y, INPUT_PULLUP);
  pinMode(END_Z, INPUT_PULLUP);
  pinMode(DISTANCE_SENSOR, INPUT);
  pinMode(FRUIT_CHOICE, INPUT);
  
  // Set Pulse Width scaling to 20%
  digitalWrite(COLOR_S0,HIGH);
  digitalWrite(COLOR_S1,LOW);

  gripper.attach(GRIPPER);

  digitalWrite(Y_ENABLE,HIGH);//Disabling the Y axis to prevent over heating
}

void loop() {
  
  //STEP 5 : Waiting For user input
  if(step ==5){
    if(!digitalRead(START_BUTTON)){
      step = 0;
    }
  }

  //STEP 0 : Reset
  if(step==0){
    Serial.println("Reseting");
    reset();
    Serial.println("Reseted");
    Serial.print("Fruit Selected : ");
    Serial.println(fruitSelected);
    step = 1;
  }
  
  //STEP 1 : Detection by scanning with Y + distance sensor
  if(step==1){
    grab(0);
    detecting();
  }
  
  //STEP 2 : Grabbing with Z
  if(step==2){
    goToObjectY();
    //grab(0);
    goToObjectZ();
    grab(1);
    step = 3;
  }
  
  //STEP 3 :  Going to color sensor and checking if color of fruit selected
  if(step==3){
    liftObject(1);
    gotoSensor();
    detectColor();
    sorting();
    step = 4;
  }
  
  //STEP 4 : Going to right Bin
  if(step==4){
    goToBin();
    liftObject(0);
    grab(0);
    reset();
    step = 5;
  }

  //STEP 6 : Transport Mode
  if(step == 6){
    transportMode();
    step = 5;
  }
  
}




void reset(){
  //Turns backwards until buttons are touched 
  grab(0);
  while(digitalRead(END_Z)){
    moveZ(1);
  }
  zPos = 0;
  digitalWrite(Y_ENABLE,LOW);//Enabling Y motor
  while(digitalRead(END_Y)){
    moveY(0);
  }
  yPos =0;
  digitalWrite(Y_ENABLE,HIGH);//Disabling the Y axis to prevent over heating
  objectStart = 0;
  objectEnd = 0;
  groundDistance = 0;
  objectRed = 0;
  objectBlue = 0;
  objectGreen = 0;
  objectFound = 0;
  bin = 1;
  digitalWrite(COLOR_LED,LOW); //Turning off LED from color sensor
  
  //Choosing the fruit
  if(analogRead(FRUIT_CHOICE) < fruit1Max){
    fruitSelected = 0;
  }
  else if(analogRead(FRUIT_CHOICE) > fruit1Max && analogRead(FRUIT_CHOICE) < fruit2Max){
    fruitSelected = 1;
  }
  else{
    fruitSelected = 2;
  }
}

void detecting(){
  digitalWrite(Y_ENABLE,LOW);//Enabling Y motor
  
  while(objectEnd == 0 && yPos < yMax){//scanning Y axis while no object found && not at max pos
    moveY(1);
    if(yPos > scanOffset){//if in scanning zone, averaging range sensor value for every 10 steps
      float averageDist  = 0;
      for( int x = 0; x<=9; x+=1){
        int instantDistance = detectDist();
        averageDist +=instantDistance;
        moveY(1);
      }
      averageDist = averageDist/10;
      if(yPos == (scanOffset + 11)){// after first averaging, distance is the ground distance
        groundDistance = averageDist;
        Serial.print("Ground Distance : ");
        Serial.println(groundDistance);
        }
    /*Serial.print(yPos);
    Serial.print(" : ");
    Serial.println(tot);*/
    if(averageDist < (groundDistance - distanceError) && !objectFound){//found start of the object
      objectStart = (yPos - sensorOffset);
      objectFound = 1;
    }
    if(averageDist < zGoal&&objectFound){//finding heighest point of the object
      zGoal = averageDist;
    }
    if(averageDist> (groundDistance - distanceError) && objectFound){//found end of the object
      objectEnd = (yPos - sensorOffset);
      Serial.print("Object height is : ");
      Serial.println((groundDistance-zGoal));
      break;
    }
    }
   
  }
  digitalWrite(Y_ENABLE,HIGH);//Disabling the Y axis to prevent over heating
  if(objectEnd == 0 || !objectFound){//if no object found, step back from 101 steps then goes to transport mode
    for(int x = 0; x<100;x+=1){
      moveY(0);
    }
    Serial.println("No object found");
    step = 6;
  }
  else{
    step = 2;
  }
}

void goToObjectY(){
  digitalWrite(Y_ENABLE,LOW);//Enabling Y motor
  yGoal = (objectEnd + objectStart)/2;// = Middle of the object
  //Serial.println(yGoal);
  bool direction = 0;
  if(yPos < yGoal){
    direction = 1; //shouldn't happen but just in case
  }
  while(abs(yPos-yGoal)>1){
    moveY(direction);
  }
  digitalWrite(Y_ENABLE,HIGH);//Disabling the Y axis to prevent over heating
}

void goToObjectZ(){
  while(zPos<(groundDistance - zSafeDistance)){//goes to the object with a safe distance
    moveZ(0);
    /*Serial.print(zPos);
    Serial.print(" : ");
    Serial.println(zGoal);*/
  }
}

void moveY(bool direction){ // 1 = forwards, 0 = backwards (adjust delays for != speed)
  digitalWrite(Y_DIR,direction);
  digitalWrite(Y_STEP, HIGH);
  delayMicroseconds(1000);
  digitalWrite(Y_STEP, LOW);
  delayMicroseconds(1000);
  if(direction){
    yPos += stepToPosY;
  }
  else{
    yPos-=stepToPosY;
  }
  //Serial.println(yPos);
  
}

void moveZ(bool direction){ // 1 = up, 0 = down (adjust delays for != speed)
  digitalWrite(Z_DIR,direction);
  digitalWrite(Z_STEP, HIGH);
  delayMicroseconds(2000);      
  digitalWrite(Z_STEP, LOW);
  delayMicroseconds(2000);
  if(direction){
    zPos -= stepToPosZ;
  }
  else{
    zPos +=stepToPosZ;
  }
}

int detectDist(){
  int analogdata = analogRead(DISTANCE_SENSOR);
  float volts = analogdata*0.0048828125;
  return 230*pow(volts,-1);//found from datasheet graph
}

void grab(bool grabbing){ //grabbing = 1 = close gripper, grabbing = 0 = open gripper
  if(grabbing){
    gripper.write(gripperAngleClosed);
  }
  else{
    gripper.write(gripperAngleOpen);
  }
}

void liftObject(bool direction){//1 = lifting, 0 = put down
  while(zPos>zObjectTransport && direction){
    moveZ(direction);
  }
  while(zPos<(groundDistance - dropDistance) && !direction){
    moveZ(direction);
  }
}

void gotoSensor(){
  int direction = ((yPos - colorReadPos) < 0);
  digitalWrite(Y_ENABLE,LOW);//Enabling Y motor
  while(abs(yPos - colorReadPos) > 1){
    moveY(direction);
  }
  yPos = 0;
  digitalWrite(Y_ENABLE,HIGH);//Disabling the Y axis to prevent over heating
}

void detectColor(){

  digitalWrite(COLOR_LED,HIGH);
  // Read Red Pulse Width
  redPW = getRedPW();
  //255 because RGB is up to 255 and inverse because pulse width 
  objectRed = map(redPW, redMin,redMax,255,0);
  objectRGB[0] = objectRed;
  // Delay to stabilize sensor
  delay(200);
  
  // Read Green Pulse Width
  greenPW = getGreenPW();
  //255 because RGB is up to 255 and inverse because pulse width 
  objectGreen = map(greenPW, greenMin,greenMax,255,0);
  objectRGB[1] = objectGreen;
  // Delay to stabilize sensor
  delay(200);
  
  // Read Blue Pulse Width
  bluePW = getBluePW();
  //255 because RGB is up to 255 and inverse because pulse width 
  objectBlue = map(bluePW, blueMin,blueMax,255,0);
  objectRGB[2] = objectBlue;
  // Delay to stabilize sensor
  delay(200);
  digitalWrite(COLOR_LED,LOW);

  Serial.print("R : ");
  Serial.print(objectRGB[0]);
  Serial.print(",G : ");
  Serial.print(objectRGB[1]);
  Serial.print(",B : ");
  Serial.println(objectRGB[2]);
}

int getRedPW() {
  // Set sensor to read Red only
  digitalWrite(COLOR_S2,LOW);
  digitalWrite(COLOR_S3,LOW);
  int PW;
  PW = pulseIn(COLOR_OUT, LOW);
  return PW;
}
int getGreenPW() {
  // Set sensor to read Green only
  digitalWrite(COLOR_S2,HIGH);
  digitalWrite(COLOR_S3,HIGH);
  int PW;
  PW = pulseIn(COLOR_OUT, LOW);
  return PW;
}
int getBluePW() {
  // Set sensor to read Blue only
  digitalWrite(COLOR_S2,LOW);
  digitalWrite(COLOR_S3,HIGH);
  int PW;
  PW = pulseIn(COLOR_OUT, LOW);
  return PW;
}

void sorting(){
  bool red =0;
  bool green = 0;
  bool blue=0;
  Serial.print("R : ");
  Serial.print(objectRGB[0]);
  Serial.print(", Goal : ");
  Serial.println(fruitColors[fruitSelected][0]);
  
  Serial.print("G : ");
  Serial.print(objectRGB[1]);
  Serial.print(", Goal : ");
  Serial.println(fruitColors[fruitSelected][1]);
  
  Serial.print("B : ");
  Serial.print(objectRGB[2]);
  Serial.print(", Goal : ");
  Serial.println(fruitColors[fruitSelected][2]);
  
  if((objectRGB[0] > (fruitColors[fruitSelected][0]-colorError)) && (objectRGB[0] < (fruitColors[fruitSelected][0]+colorError))){ //check if red value of object is in the interval
    red = 1;
  }
  if((objectRGB[1] > (fruitColors[fruitSelected][1]-colorError)) && (objectRGB[1] < (fruitColors[fruitSelected][1]+colorError))){//check if green value of object is in the interval
    green = 1;
  }
  if((objectRGB[2] > (fruitColors[fruitSelected][2]-colorError)) && (objectRGB[2] < (fruitColors[fruitSelected][2]+colorError))){//check if blue value of object is in the interval
    blue = 1;
  }
  if(red && green && blue){//if R,G and B are in interval, OK zone
    Serial.println("right color");
    bin = 0;
  }
}

void goToBin(){
  //Serial.println("going to bin");
  digitalWrite(Y_ENABLE,LOW);//Enabling Y motor
  while(abs(yPos-yBins[bin])>1){
    moveY(1);
  }
  digitalWrite(Y_ENABLE,HIGH);//Disabling the Y axis to prevent over heating
}

void transportMode(){ // position of rest for transport
  digitalWrite(Y_ENABLE,LOW);//Enabling Y motor
  while(abs(yPos-yMax)>1){
    moveY(1);
  }
  digitalWrite(Y_ENABLE,HIGH);//Disabling the Y axis to prevent over heating
  int distance = detectDist();
  while(zPos<(distance - dropDistance)){
    moveZ(0);
  }
}
